#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#
require 'zlib'

module Gonzui
  class IncompatibleDBError < GonzuiError; end
  DB_VERSION = "1"

  include Util

  class DBM
    def self.open (config, read_only = false)
      File.mkpath(config.db_directory) unless read_only

      dbm_class = BDBDBM # to be pluggable
      dbm = dbm_class.new(config, read_only)
      if block_given?
        begin
          yield(dbm)
        ensure
          dbm.close
        end
      else
        return dbm
      end
    end
  end

  class DBMError < GonzuiError; end
  class AbstractDBM
    def initialize (config, read_only = false)
      raise "#{config.db_directory}: No such directory" unless 
        File.directory?(config.db_directory)
      @config = config

      db_table = [ 
        [:langid_lang,       AutoPack::IdAndString,     false],
        [:langid_labbr,      AutoPack::IdAndString,     false],
        [:labbr_langid,      AutoPack::StringAndId,     false],
        [:seq,               AutoPack::StringAndNumber, false],
        [:stat,              AutoPack::StringAndNumber, false],
        [:pkg_pkgid,         AutoPack::StringAndId,     false],
        [:pkgid_pkg,         AutoPack::IdAndString,     false],
        [:pkgid_pathids,     AutoPack::IdAndId,         true],
        [:pkgid_langids,     AutoPack::IdAndString,     true],
        [:path_pathid,       AutoPack::StringAndId,     false],
        [:pathid_digest,     AutoPack::IdAndString,     false], # multi-value
        [:pathid_info,       AutoPack::IdAndString,     false],
        [:pathid_content,    AutoPack::IdAndString,     false],
        [:pathid_path,       AutoPack::IdAndString,     false],
        [:pathid_pkgid,      AutoPack::IdAndId,         false],
        [:pathid_wordids,    AutoPack::IdAndString,     false], # multi-value
        [:word_wordid,       AutoPack::StringAndId,     false],
        [:wordid_pkgids,     AutoPack::IdAndId,         true],
        [:wordid_word,       AutoPack::IdAndString,     false],
        [:wordpkgid_pathids, AutoPack::StringAndId,     true],
        [:wordpathid_info,   AutoPack::StringAndString, false], # multi-value
      ]

      validate_db_version
      @db_opened = Hash.new
      db_table.each {|db_name, pack_adaptor, dupsort|
        open_db(db_name, pack_adaptor, dupsort)
      }
      save_db_version unless read_only
      @opened = true
      @word_cache = Hash.new
      @funcfreq_cache = Hash.new
    end

    private
    def collect_all_results (word_id)
      results = []
      if word_id
        self.wordid_pkgids.each_dup(word_id) {|k, package_id|
          word_package_id = word_id + package_id
          self.wordpkgid_pathids.each_dup(word_package_id) {|k, path_id|
            results.concat(get_all_word_info(word_id, path_id))
          }
        }
      end
      return results
    end

    def db_exist?
      return false unless File.directory?(@config.db_directory)
      entries = Dir.entries_without_dots(@config.db_directory)
      # filter out file names like __db.001.
      entries = entries.find_all {|entry| not /^__/.match(entry) }
      if entries.empty?
        return false
      else
        return true
      end
    end

    def decrease_counter (key, step = 1)
      value = get_counter(key) - step
      raise DBMError.new("counter #{key} becomes minus") if value < 0
      self.stat[key.to_s] = value
    end


    def do_open_db (name, dupsort)
      raise NotImplementedError.new
    end

    FuncFreqPrefix = "ffreq"
    FuncFreqSuffix = "rev"
    def each_freq_db_name
      by_name_regex = /^#{FuncFreqPrefix}_[a-z]+$/
      Dir.entries_without_dots(@config.db_directory).each {|entry|
        if by_name_regex.match(entry)
          by_name = entry
          by_freq = (entry + "_" + FuncFreqSuffix)
          yield(by_name, by_freq)
        end
      }
    end

    def open_all_funcfreq_dbs
      each_freq_db_name {|by_name, by_freq|
        open_db(by_name, AutoPack::StringAndNumber, false)
        open_db(by_freq, AutoPack::NumberAndString, true)
      }
    end

    def open_freq_db (lang_abbrev)
      by_name = sprintf("%s_%s", FuncFreqPrefix, lang_abbrev).intern
      by_freq = sprintf("%s_%s_%s", FuncFreqPrefix, 
                        lang_abbrev, FuncFreqSuffix).intern
      db_by_name = open_db(by_name, AutoPack::StringAndNumber, false)
      db_by_freq = open_db(by_freq, AutoPack::NumberAndString, true)
      return db_by_name, db_by_freq
    end

    def open_db (db_name, pack_adaptor, dupsort)
      return if @db_opened.include?(db_name)
      db = do_open_db(db_name, dupsort)
      db.extend(pack_adaptor)
      @db_opened[db_name] = db

      name = "@" + db_name.to_s
      instance_variable_set(name, db)
      self.class.class_eval { 
        attr_reader name.delete("@")
      }
      return db
    end

    def save_db_version
      db = do_open_db(:version, false)
      db["version"] = DB_VERSION
      db.close
    end
      
    def validate_db_version
      return unless db_exist?
      version = "unknown"
      begin
        db = do_open_db(:version, false)
        version = db["version"]
        db.close
      rescue BDB::Fatal
      end
      if version != DB_VERSION
        m = sprintf("DB format is incomatible (version %s expected but %s)",
                    DB_VERSION, version)
        raise IncompatibleDBError.new(m)
      end
    end

    def verify_stat_integrity
      assert_equal(self.nlanguages, 
                   self.langid_lang.length, 
                   self.langid_labbr.length, 
                   self.labbr_langid.length)
      assert_equal(self.npackages,
                   self.pkgid_pkg.length, 
                   self.pkg_pkgid.length)
      assert_equal(self.nfiles,
                   self.path_pathid.length, 
                   self.pathid_path.length,
                   self.pathid_content.length,
                   self.pathid_info.length)
      assert_equal(self.nwords,  
                   self.word_wordid.length)
      nlines = 0
      self.pathid_info.each_key {|path_id|
        nlines += get_file_info(path_id).nlines if source_file?(path_id)
      }
      assert_equal(self.nlines, nlines)
    end

    def verify_seq_integrity
      table = [
        ["last_lang_id",    :langid_lang],
        ["last_package_id", :pkgid_pkg],
        ["last_path_id",    :pathid_path],
        ["last_word_id",    :wordid_word]
      ]
      table.each {|id_name, db_name|
        id = self.send(db_name).get_last_key
        idx = id ? id_to_index(id) : 0
        assert(idx <= (self.seq[id_name] or 0))
      }
    end

    def verify_freq_integrity
      open_all_funcfreq_dbs

      by_names = []
      by_freqs = []
      each_freq_db_name {|by_name, by_freq|
        by_names.push(by_name)
        by_freqs.push(by_freq)
      }
      assert_equal(by_names.length, by_freqs.length)

      by_names.sort!
      by_freqs.sort!
      by_names.length.times {|i|
        by_name = self.send(by_names[i])
        by_freq = self.send(by_freqs[i])
        assert_equal(by_name.length, by_freq.length)
      }
    end

    public
    def close
      raise DBMError.new("dbm is already closed") unless @opened
      @db_opened.each {|name, db| 
        db.close
      }
      @opened = false
    end

    def word_cache
      @word_cache
    end

    def consistent?
      verify_stat_integrity
      verify_seq_integrity
      verify_freq_integrity
      return true
    end

    def decrease_counter (key, step = 1)
      value = self.get_counter(key) - step
      raise DBMError.new("counter #{key} becomes minus") if value < 0
      self.stat[key.to_s] = value
    end

    def each_db_name
      open_all_funcfreq_dbs
      @db_opened.each_key {|db_name| yield(db_name.to_s) }
    end

    def each_language
      self.langid_lang.each {|lang_id, lang_name|
        lang_abbrev = get_lang_abbrev(lang_id)
        yield(lang_id, lang_abbrev, lang_name)
      }
    end

    def each_package_name
      self.pkgid_pkg.each_value {|value| yield(value) }
    end

    def each_word (&block)
      self.word_wordid.each_key {|word| yield(word) }
    end

    def find_all (pattern)
      word_id = self.word_wordid[pattern]
      results = collect_all_results(word_id)
      return results
    end

    def find_all_by_prefix (pattern)
      raise NotImplementedError.new("should be implemented in a sub class")
    end

    def find_all_by_regex (pattern)
      regex = Regexp.new(pattern)
      results = []
      self.word_wordid.each {|word, word_id|
        if regex.match(word)
          results.concat(collect_all_results(word_id))
        end
      }
      return results
    end

    def find_word_info (word_id, path_id)
      dump = self.wordpathid_info[word_id + path_id]
      DeltaDumper.undump(WordInfo, dump) {|data|
        yield(WordInfo.new(word_id, path_id, *data))
      }
    end

    def get_all_word_info (word_id, path_id)
      dump = self.wordpathid_info[word_id + path_id]
      DeltaDumper.undump(WordInfo, dump).map {|data|
        WordInfo.new(word_id, path_id, *data)
      }
    end

    def get_counter (key)
      self.stat[key.to_s] or 0
    end

    def get_file_content (path_id)
      content = self.pathid_content[path_id]
      content = Zlib::Inflate.inflate(content)
      return content
    end

    def get_file_info (path_id)
      dump = self.pathid_info[path_id]
      assert(dump)
      return FileInfo.load(dump)
    end

    def get_function_popularity (lang_abbrev, n)
      funcfreq = make_funcfreq(lang_abbrev)
      data = funcfreq.tail(n).reverse
      return data
    end

    def get_digest (path_id)
      dump = self.pathid_digest[path_id]
      DeltaDumper.undump(DigestInfo, dump).map {|data|
        DigestInfo.new(*data)
      }
    end

    def get_lang_abbrev (lang_id)
      self.langid_labbr[lang_id]
    end

    def get_lang_id (lang_abbrev)
      self.labbr_langid[lang_abbrev]
    end

    def get_lang_id_from_path_id (path_id)
      get_file_info(path_id).lang_id
    end

    def get_lang_ids_from_package_id (package_id)
      self.pkgid_langids.duplicates(package_id)
    end

    def get_lang_name (lang_id)
      self.langid_lang[lang_id]
    end

    def get_nsources_by_lang_id (lang_id)
      lang_abbrev = self.get_lang_abbrev(lang_id)
      key = make_nsources_key(lang_abbrev)
      return self.get_counter(key)
    end

    def get_package_id (package_name)
      self.pkg_pkgid[package_name]
    end

    def get_package_id_from_path_id (path_id)
      self.pathid_pkgid[path_id]
    end

    def get_package_ids (word_id)
      self.wordid_pkgids.duplicates(word_id)
    end

    def get_package_name (package_id)
      self.pkgid_pkg[package_id]
    end

    def get_path (path_id)
      self.pathid_path[path_id]
    end

    def get_path_id (path)
      self.path_pathid[path]
    end

    def get_path_ids (package_id)
      self.pkgid_pathids.duplicates(package_id)
    end

    def get_path_ids_from_word_and_package_id (word_id, package_id)
      word_package_id = word_id + package_id
      self.wordpkgid_pathids.duplicates(word_package_id)
    end

    def get_word (word_id)
      self.wordid_word[word_id]
    end

    def get_word_id (word)
      self.word_wordid[word]
    end

    def get_word_ids (path_id)
      self.pathid_wordids_load(path_id)
    end

    def has_lang_id? (lang_id)
      self.langid_lang.has_key?(lang_id)
    end

    def has_lang_abbrev? (lang_abbrev)
      self.labbr_langid.has_key?(lang_abbrev)
    end

    def has_package? (package_name)
      self.pkg_pkgid.include?(package_name)
    end

    def has_word? (word)
      wordid = self.word_wordid[word]
      if wordid
        return true
      else
        return false
      end
    end

    def increase_counter (key, step = 1)
      self.stat[key.to_s] = get_counter(key) + step
    end

    def id_to_index (id)
      id.unpack("N").first
    end

    def index_to_id (idx)
      [idx].pack("N")
    end

    def make_funcfreq (lang_abbrev)
      unless @funcfreq_cache[lang_abbrev]
        db_by_name, db_by_freq = open_freq_db(lang_abbrev)
        funcfreq = FrequencyCounter.new(db_by_name, db_by_freq)
        @funcfreq_cache[lang_abbrev] = funcfreq
      end
      return @funcfreq_cache[lang_abbrev]
    end

    def make_new_id (id_name)
      last_id_key = "last_" << id_name.to_s
      last_idx = self.seq[last_id_key]
      new_idx = (last_idx or -1) + 1
      self.seq[last_id_key] = new_idx
      return index_to_id(new_idx)
    end

    def make_nsources_key (lang_abbrev)
      ("nsources_" + lang_abbrev).intern
    end

    def nfiles
      get_counter(:nfiles)
    end

    def nlines
      get_counter(:nlines)
    end

    def nlanguages
      get_counter(:nlanguages)
    end

    def npackages
      get_counter(:npackages)
    end

    def nwords
      get_counter(:nwords)
    end

    def nsources
      get_counter(:nsources)
    end

    def nfiles_in_package (package_name)
      package_id = self.get_package_id(package_name)
      self.pkgid_pathids.duplicates(package_id).length
    end

    def set_file_content (path_id, content)
      self.pathid_content[path_id] = Zlib::Deflate.deflate(content)
    end

    def source_file? (path_id)
      lang_id = get_lang_id_from_path_id(path_id)
      return has_lang_id?(lang_id)
    end

    def pathid_wordids_dump (path_id, ids)
      self.pathid_wordids[path_id] = ids.join.unpack("N*").pack("w*")
    end

    def pathid_wordids_load (path_id)
      self.pathid_wordids[path_id].unpack("w*").map {|x| [x].pack("N") }
    end
  end

  class FrequencyCounter
    def initialize (db_by_name, db_by_freq)
      @db_by_name = db_by_name
      @db_by_freq = db_by_freq
    end

    def decrease (name, step = 1)
      increase(name, - step)
    end

    def increase (name, step = 1)
      prev = @db_by_name[name]
      freq = (prev or 0) + step
      assert(freq >= 0)
      @db_by_freq.delete_both(prev, name) if prev
      if freq > 0
        @db_by_name[name] = freq
        @db_by_freq[freq] = name
      elsif freq == 0
        @db_by_name.delete(name)
      else
        assert_not_reached
      end
    end

    def tail (n)
      @db_by_freq.tail(n)
    end
  end
end

