#
# gonzui - a source code search engine.
#
# Copyright (C) 2004-2005 Satoru Takabayashi <satoru@namazu.org> 
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#

require "gonzui/progressbar"

module Gonzui
  include Util

  class ImporterError < GonzuiError; end
  class PackageAlreadyExists < GonzuiError; end

  class Importer
    def initialize (config, options = {})
      @start_time = Time.now
      @dbm = DBM.open(config)
      @config = config
      @nfiles = 0
      @npackages = 0
      @show_progress = options[:show_progress]
      @last_package = nil
    end
    attr_reader :last_package

    private
    def make_progress_bar (title, total, reversed = false)
      klass = if @show_progress
                if reversed
                  ReversedProgressBar
                else
                  ProgressBar
                end
              else
                NullObject
              end
      return klass.new(title, total)
    end

    def add_package (package)
      raise PackageAlreadyExists.new("#{package.name}: already exists") if 
        @dbm.has_package?(package.name)

      verbose = @config.verbose
      pbar = make_progress_bar(package.name, package.nfiles) unless verbose
      package.each {|real_path, normalized_path|
        protect_from_signals {
          content = File.read(real_path)
          mtime = File.mtime(real_path)
          indexer = Indexer.new(@config, @dbm, normalized_path, content, mtime)
          indexer.index
        }
        pbar.inc unless verbose
      }
      pbar.finish unless verbose

      @npackages += 1
      @nfiles    += package.nfiles
      @last_package = package
    end

    def add_common (worker, method)
      begin
        directory = worker.send(method)
        add_directory(directory)
      ensure
        worker.clean
      end
    end

    public
    def add_directory (directory)
      package = Package.new(directory)
      add_package(package)
    end

    def add_archive (file_name)
      extractor = Extractor.new(@config, file_name)
      add_common(extractor, :extract)
    end

    def add_by_apt (package_name)
      aptget = AptGet.create(@config, package_name)
      add_common(aptget, :get)
    end

    def task_name
      "imported"
    end

    def summary
      elapsed = Time.now - @start_time
      format = "%d files of %d packages %s in %.2f sec. (%.2f files / sec.)\n"
      ret = sprintf(format, @nfiles, @npackages, task_name, elapsed,
                    @nfiles / elapsed)
      if @config.verbose
        stat = Indexer.statistics
        ret += "\n" + stat unless stat.empty?
      end
      return ret
    end

    def finish
      @dbm.close
    end
  end

  class RemoverError < GonzuiError; end

  class Remover < Importer
    def remove_package (package_name)
      raise RemoverError.new("#{package_name}: package not found") unless
        @dbm.has_package?(package_name)
      nfiles = @dbm.nfiles_in_package(package_name)

      pbar = ProgressBar.new(package_name, nfiles)
      package_id = @dbm.get_package_id(package_name)
      @dbm.get_path_ids(package_id).each {|path_id|
        normalized_path = @dbm.get_path(path_id)
        protect_from_signals {
          deindexer = Deindexer.new(@config, @dbm, normalized_path)
          deindexer.deindex
        }
        pbar.inc
      }
      pbar.finish

      @npackages += 1
      @nfiles += nfiles
    end

    def task_name
      "removed"
    end
  end
end
