require 'langscan/java/java.so'
require 'langscan/_common'

# define Array#rindex {} for 1.8 or older 
begin
  [].rindex {}
rescue ArgumentError
  class Array
    alias clang_original_rindex rindex
    def rindex(*args)
      if block_given?
        (length-1).downto(0) {|i|
          return i if yield self[i]
        }
        nil
      else
        clang_original_rindex(*args)
      end
    end
  end
end

module LangScan
  module Java
    def lang_name
      "Java"
    end
    module_function :lang_name

    def lang_abbrev
      "java"
    end
    module_function :lang_abbrev

    def extnames
      [".java"]
    end
    module_function :extnames

    # LangScan::Java.scan iterates over Java program.
    # It yields for each element which is interested by gonzui. 
    #
    # _input_ should be a String or a IO.
    def scan(input, &block)
      stack = []
      last_pair = nil
      last_token = nil
      fragments = {}
      each_fragment(input) {|t|
        if t.type == :space || t.type == :comment
          yield t
          next
        end
        fragments[t.beg_byteno] = t
        if last_pair
          last_pair.after_close_token = t
          last_pair.detected {|f|
            fragments.delete f.beg_byteno
            yield f
          }
          last_pair = nil
        end
        if !stack.empty? &&
           stack.last.after_open_tokens.length < 2 &&
           (stack.last.after_open_tokens.last || stack.last.open_token) == last_token
          stack.last.after_open_tokens << t
        end
        case t.type
        when :preproc_beg
          stack.push Pair.new(stack.last, :preproc, last_token, t)
        when :preproc_end
          if level = stack.rindex {|pair| pair.type == :preproc }
            last_pair = stack[level]
            last_pair.before_close_token = last_token
            last_pair.close_token = t
            stack[level..-1] = []
          end
        when :punct
          case t.text
          when '('
            stack.push Pair.new(stack.last, :paren, last_token, t)
            if 1 < stack.length && stack[0].type == :paren
              # function type in formal parameters, maybe.
              stack.last.maybe_function_open_paren = false
            end
          when ')'
            if level = stack.rindex {|pair| pair.type == :paren }
              last_pair = stack[level]
              last_pair.before_close_token = last_token
              last_pair.close_token = t
              stack[level..-1] = []
            end
          when '{'
            stack.push Pair.new(stack.last, :brace, last_token, t)
          when '}'
            if level = stack.rindex {|pair| pair.type == :brace }
              last_pair = stack[level]
              last_pair.before_close_token = last_token
              last_pair.close_token = t
              stack[level..-1] = []
            end
          when '['
            stack.push Pair.new(stack.last, :bracket, last_token, t)
          when ']'
            if level = stack.rindex {|pair| pair.type == :bracket }
              last_pair = stack[level]
              last_pair.before_close_token = last_token
              last_pair.close_token = t
              stack[level..-1] = []
            end
          when ','
            stack.last.comma_tokens << t if !stack.empty?
          end
        end
        last_token = t
      }
      if last_pair
        last_pair.detected {|f|
          fragments.delete f.beg_byteno
          yield f
        }
      end
      fragments.each_value {|f|
        yield f
      }
    end
    module_function :scan

    # LangScan::Java.each_fragment iterates over Java-language fragments in _input_.
    # The fragments contains tokens and inter-token spaces and comments.
    #
    # _input_ can be specified as a String or a IO.
    # If a String is specified, the String itself is assumed as a Java-program.
    # If a IO is specified, the content of the IO is assumed as a Java-program.
    def each_fragment(input) # :yields: token
      begin
        tokenizer = Tokenizer.new(input)
        while token_info = tokenizer.get_token
          type, text, beg_lineno, beg_columnno, beg_byteno, end_lineno, end_columnno, end_byteno = token_info
          token = Fragment.new(type, text, beg_lineno, beg_byteno)
          if token.type == :ident
            if TypesHash[token.text]
              token.type = :type
            elsif KeywordsHash[token.text]
              token.type = :keyword 
            end
          end
          yield token
        end
      ensure
        tokenizer.close
      end
    end
    module_function :each_fragment

    # LangScan::Java.each_token iterates over Java-language tokens in _input_.
    #
    # _input_ can be specified as a String or a IO.
    # If a String is specified, the String itself is assumed as a Java-program.
    # If a IO is specified, the content of the IO is assumed as a Java-program.
    def each_token(input) # :yields: token
      begin
        tokenizer = Tokenizer.new(input)
        while token_info = tokenizer.get_token
          type, text, beg_lineno, beg_columnno, beg_byteno, end_lineno, end_columnno, end_byteno = token_info
          if type != :space && type != :comment
            yield Fragment.new(type, text, beg_lineno, beg_byteno)
          end
        end
      ensure
        tokenizer.close
      end
    end
    module_function :each_token

    Keywords = %w(
      abstract boolean break byte case catch char class const continue default
      do double else extends final finally float for goto if implements import
      int interface long native new package private protected public return
      short static strictfp super switch synchronized this throw throws
      transient try void volatile while true false null
    )
    KeywordsHash = {}
    Keywords.each {|k| KeywordsHash[k] = k }

    Types = %w(boolean byte char double float int long short void)
    TypesHash = {}
    Types.each {|k| TypesHash[k] = k }

    # :stopdoc:
    class Pair
      def initialize(outer, type, before_open_token, open_token)
        @outer = outer
        @type = type
        @before_open_token = before_open_token
        @open_token = open_token
        @after_open_tokens = []
        @comma_tokens = []
        @before_close_token = nil
        @close_token = nil
        @after_close_token = nil
        @maybe_function_open_paren = true
      end
      attr_reader :outer, :type, :before_open_token, :open_token, :comma_tokens
      attr_accessor :after_open_tokens, :before_close_token, :close_token, :after_close_token
      attr_accessor :maybe_function_open_paren

      def pretty_print_instance_variables
        [:@type,
         :@before_open_token, :@open_token, :@after_open_tokens,
         :@comma_tokens,
         :@before_close_token, :@close_token, :@after_close_token,
         :@maybe_function_open_paren]
      end

      def detected
        if @type == :paren &&
           @before_open_token && @before_open_token.type == :ident &&
           @maybe_function_open_paren
          num_args = @comma_tokens.length+1 
          num_args = 0 if @open_token == @before_close_token
          if !KeywordsHash[@before_open_token.text]
            if @outer &&
               @outer.type == :preproc &&
               1 < @outer.after_open_tokens.length &&
               @outer.after_open_tokens[0].type == :ident &&
               @outer.after_open_tokens[0].text == 'define' &&
               @outer.after_open_tokens[1] == @before_open_token
              if @before_open_token.end_byteno == @open_token.beg_byteno
                @before_open_token.type = :fundef
                yield @before_open_token
              end
            elsif !@outer || !@outer.outer
              if @after_close_token &&
                 @after_close_token.type == :punct &&
                 @after_close_token.text == '{'
                @before_open_token.type = :fundef
                yield @before_open_token
              else
                @before_open_token.type = :funcall
                yield @before_open_token
              end
            else
              @before_open_token.type = :funcall
              yield @before_open_token
            end
          end
        end
      end
    end
    # :startdoc:

    # LangScan::Java.each_function searches function definition and function call in _input_.
    #
    # _input_ should be a String or a IO.
    #
    # LangScan::Java.each_function yields a LangScan::FunDef or a LangScan::FunCall.
    #
    def each_function(input, &block) # :yields: fundef or funcall
      stack = []
      last_pair = nil
      last_token = nil
      each_token(input) {|t|
        if last_pair
          last_pair.after_close_token = t
          last_pair.detected(&block)
          last_pair = nil
        end
        if !stack.empty? &&
           stack.last.after_open_tokens.length < 2 &&
           (stack.last.after_open_tokens.last || stack.last.open_token) == last_token
          stack.last.after_open_tokens << t
        end
        case t.type
        when :preproc_beg
          stack.push Pair.new(stack.last, :preproc, last_token, t)
        when :preproc_end
          if level = stack.rindex {|pair| pair.type == :preproc }
            last_pair = stack[level]
            last_pair.before_close_token = last_token
            last_pair.close_token = t
            stack[level..-1] = []
          end
        when :punct
          case t.text
          when '('
            stack.push Pair.new(stack.last, :paren, last_token, t)
            if 1 < stack.length && stack[0].type == :paren
              # function type in formal parameters, maybe.
              stack.last.maybe_function_open_paren = false
            end
          when ')'
            if level = stack.rindex {|pair| pair.type == :paren }
              last_pair = stack[level]
              last_pair.before_close_token = last_token
              last_pair.close_token = t
              stack[level..-1] = []
            end
          when '{'
            stack.push Pair.new(stack.last, :brace, last_token, t)
          when '}'
            if level = stack.rindex {|pair| pair.type == :brace }
              last_pair = stack[level]
              last_pair.before_close_token = last_token
              last_pair.close_token = t
              stack[level..-1] = []
            end
          when '['
            stack.push Pair.new(stack.last, :bracket, last_token, t)
          when ']'
            if level = stack.rindex {|pair| pair.type == :bracket }
              last_pair = stack[level]
              last_pair.before_close_token = last_token
              last_pair.close_token = t
              stack[level..-1] = []
            end
          when ','
            stack.last.comma_tokens << t if !stack.empty?
          end
        end
        last_token = t
      }
      last_pair.detected(&block) if last_pair
    end
    module_function :each_function
  end

end

if LangScan.respond_to?(:register)
  LangScan.register(LangScan::Java)
end
