require 'langscan/php/php.so'
require 'langscan/_common'

module LangScan
  module PHP
    def lang_name
      "PHP"
    end
    module_function :lang_name

    def lang_abbrev
      "php"
    end
    module_function :lang_abbrev

    def extnames
      [".php"]
    end
    module_function :extnames

    # LangScan::PHP.scan iterates over PHP program.
    # It yields for each element which is interested by gonzui. 
    #
    # _input_ should be a String or a IO.
    def scan(input, &block)
      last_token = nil
      each_fragment(input) {|t|
        if t.type == :space
          yield t
          next
        end
        if last_token
          if t.type == :ident and last_token.type == :keyword
            case last_token.text
            when 'function'
              t.type = :fundef
            when 'class'
              t.type = :classdef
            end
          elsif t.type == :punct and t.text == '(' and last_token.type == :ident
            last_token.type = :funcall
          end
          yield last_token
          last_token = nil
        end
        if t.type == :ident or t.type == :keyword
          last_token = t
        else
          yield t
        end
      }
      if last_token
        yield last_token
      end
    end
    module_function :scan

    # LangScan::PHP.each_fragment iterates over PHP-language fragments in _input_.
    # The fragments contains tokens and inter-token spaces and comments.
    #
    # _input_ can be specified as a String or a IO.
    # If a String is specified, the String itself is assumed as a PHP-program.
    # If a IO is specified, the content of the IO is assumed as a PHP-program.
    def each_fragment(input) # :yields: token
      begin
        tokenizer = Tokenizer.new(input)
        while token_info = tokenizer.get_token
          type, text, beg_lineno, beg_columnno, beg_byteno, end_lineno, end_columnno, end_byteno = token_info
          token = Fragment.new(type, text, beg_lineno, beg_byteno)
          if token.type == :ident
            if KeywordsHash[token.text]
              token.type = :keyword 
            end
          end
          yield token
        end
      ensure
        tokenizer.close
      end
    end
    module_function :each_fragment

    Keywords = %w(
		and      or      xor     __FILE__        exception       php_user_filter
		__LINE__        array   as      break   case
		class   const   continue        declare         default
		die     do      echo    else    elseif
		empty   enddeclare      endfor  endforeach      endif
		endswitch       endwhile        eval    exit    extends
		for     foreach         function        global  if
		include         include_once    isset   list    new
		print   require         require_once    return  static
		switch  unset   use     var     while
		__FUNCTION__    __CLASS__       __METHOD__      final   php_user_filter
		interface       implements      extends         public          private
		protected       abstract        clone   try     catch
		throw   cfunction       old_function
		    )
    KeywordsHash = {}
    Keywords.each {|k| KeywordsHash[k] = k }

  end
end

if LangScan.respond_to?(:register)
  LangScan.register(LangScan::PHP)
end

